/*
 * SPI2CF (c) 2006, Till Harbaum.
 * All rights reserved.
 *
 * This file is distributed under the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part SPI2CF port of the uIP TCP/IP stack.
 */

/*
  cf.c
*/

#include "uip-conf.h"

#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>

#include "spi.h"
#include "cf.h"

/* physically reset the card */
void cf_reset(void) {

  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(0x8000);

  _delay_ms(10); 

  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(0x0000);

  _delay_ms(200); 
}

void cf_get_id(void) {
  uint16_t id;

  spi_start();
  spi_write_byte(CMD_BYTE(GET_ID));
  id = spi_read_word();
  printf("SPI2CF chip id: %02x.%02x\n", 
	 id>>8, id&0xff);
}

void cf_set_addr(uint16_t addr) {

  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(addr);
}

void cf_reg_read(uint16_t addr, char *dst, uint16_t bytes) {
  unsigned char result;

  // set address
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(addr);

  // and issue register read command
  spi_start();
  spi_write_byte(CMD_BYTE(REG_READ));
  while(bytes-- > 1)
    *dst++ = spi_read_byte();

  // read last byte using special mechanism
  spi_start();
  *dst++ = spi_cycle_byte(CMD_BYTE(GET_ID));
}

// read two bytes but make sure that the last byte does
// not cause the transfer of another one
uint16_t cf_read_last_word(void) {
  uint16_t result;
  unsigned char *p = (unsigned char*)&result;

  *p++ = spi_read_byte();

  // read last byte using special mechanism
  spi_start();
  *p++ = spi_cycle_byte(CMD_BYTE(GET_ID));

  return result;
}

unsigned char cf_reg_read_byte(uint16_t addr) {
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(addr);

  spi_start();
  spi_write_byte(CMD_BYTE(REG_READ));
  return spi_read_byte();
}

uint16_t cf_reg_read_word(uint16_t addr) {
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(CF_ADDR_BINC | addr);

  spi_start();
  spi_write_byte(CMD_BYTE(REG_READ));
  return cf_read_last_word();
}

void cf_reg_write_byte(uint16_t addr, unsigned char byte) {
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(CF_ADDR_BINC | addr);

  spi_start();
  spi_write_byte(CMD_BYTE(REG_WRITE));
  spi_write_byte(byte);
  spi_write_byte(0xff);  // dummy byte will never be written
}

uint16_t cf_io_read_word(uint16_t addr) {
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(CF_ADDR_BINC + addr);

  spi_start();
  spi_write_byte(CMD_BYTE(IO_READ));
  return cf_read_last_word();
}

void cf_io_write_word(uint16_t addr, uint16_t val) {
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(CF_ADDR_BINC + addr);

  spi_start();
  spi_write_byte(CMD_BYTE(IO_WRITE));
  spi_write_word(val);
  spi_write_byte(0xff);  // dummy byte will never be written
}

void cf_io_read(uint16_t addr, uint16_t *dst, uint16_t words) {
  unsigned char result, *dbyte = (unsigned char*)dst;

  // set address
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(CF_ADDR_TOGGLE | addr);

  // and issue io read command
  spi_start();
  spi_write_byte(CMD_BYTE(IO_READ));
  while(words-- > 1) {
    *dbyte++ = spi_read_byte();
    *dbyte++ = spi_read_byte();
  }

  // read last byte using special mechanism
  *dbyte++ = spi_read_byte();
  spi_start();
  *dbyte++ = spi_cycle_byte(CMD_BYTE(GET_ID));
}

void cf_io_write(uint16_t addr, uint16_t *src, uint16_t words) {
  unsigned char result, *sbyte = (unsigned char*)src;

  // set address
  spi_start();
  spi_write_byte(CMD_BYTE(SET_ADDR));
  spi_write_word(CF_ADDR_TOGGLE | addr);

  // and issue io read command
  spi_start();
  spi_write_byte(CMD_BYTE(IO_WRITE));
  while(words--) {
    spi_write_byte(*sbyte++);
    spi_write_byte(*sbyte++);
  }

  spi_write_byte(0xff);  // dummy byte will never be written
}

